#' ## Parsing the LVMC Case Summary Report
#' 
#' Finally got the time to put together the omnibus project I had mentioned in a previous entry. Doing this in R for two reasons: first, because I'm rather not get rusty with R (Anki helps), and second, to use the XML library. 
#' 
#' ## Preliminaries and scraping
#' 
#' Here we load the library and select all tables nested within a parent table cell in the tree.
#' 
## @knitr libraries, warning=FALSE,  echo=TRUE, message=FALSE
library(XML)
library("data.table", lib.loc="/home/james/R")
library("plyr", lib.loc="/home/james/R")
library("ggplot2", lib.loc="/home/james/R")
library("stringr", lib.loc="/home/james/R")
library("xtable", lib.loc="/home/james/R")

#' 
#' 
#' We can make use of the script we used, and include it in its entirety thanks to the knitr python engine.
#' 
## @knitr get-files, eval=FALSE, engine='python'
## import gzip
## import sys
## from selenium import webdriver
## from selenium.webdriver.common.keys import Keys
## 
## def get_lvmc_case_report(caseNumber):
##     driver = webdriver.Firefox()
##     driver.get('https://secure2.lasvegasnevada.gov/defendantreport/Default.aspx')
##     casenum = driver.find_element_by_name('txt_CaseNo')
##     casenum.send_keys(caseNumber)
##     casenum.send_keys(Keys.RETURN)
##     driver.implicitly_wait("3")
## 
##     for handle in driver.window_handles:
##         driver.switch_to_window(handle)
##         if driver.current_url == 'https://secure2.lasvegasnevada.gov/defendantreport/report.aspx':
##             filename = caseNumber + '.html.gz'
##             outfile = gzip.open(filename, 'wb')
##             outfile.write(driver.page_source)
##             outfile.close()
##             print "Wrote %s" % filename
##     driver.quit()
## 
## cases = ['C' + str(x) + 'A' for x in range(1000001, 1099263, 1)]
## 
## for case in cases:
##     print case
##     get_lvmc_case_report(case)
##     # The time it takes to open a browser window is enough,
##     # no need to sleep
##     # sleep(randint(0,1))

#' 
#' 
#' See the original post for documentation. 
#' 
## @knitr count-files, cache=TRUE
results.list <- list.files(path = "/home/james/Dropbox/legal/scraping/lvmc", 
                           pattern = "*.html.gz",
                           recursive = TRUE,
                           full.names = TRUE)
names(results.list) <- gsub("/home/james/Dropbox/legal/scraping/lvmc/\\d\\d/(C\\d+A).html.gz", "\\1", results.list)
head(results.list)          
tail(results.list)

#' 
#' 
#' It wasn't necessary to download local
#' copies of our files; we could have constructed and fed the URLs directly into
#' the `XML` parser, we'd just replace the following strings with the appropriate
#' URL generator:
#' 
## @knitr tree-parse-def, cache=TRUE

TreeParse <-  function(x) {
  tree <- htmlTreeParse(x, useInternalNodes=TRUE, asText=FALSE)
  srctbl  <-  getNodeSet(tree, "//span/div/table/tbody/tr/td/table")
  return(srctbl)
  }
   

#' 
#' 
#' [Xiao Nan](http://cos.name/) has a nice slide deck on web scraping with R available [here](http://cos.name/wp-content/uploads/2013/05/Web-Scraping-with-R-XiaoNan.pdf).
#' 
#' ## Case Identifiers
#' 
#' We also create a class for the LVMC's date format:
## @knitr caseDate-class
setClass("caseDate")
setAs("character","caseDate", function(from) as.POSIXct(from, format = "%m/%d/%Y %I:%M %p"))

#' 
#' 
#' And for the ID table, we need to split table cells by colon.
## @knitr pairVectors-def
pairVectors <- function(x) {
  # Splits cells from the identifiers table of LVMC case reports into variables
  #
  # Args:
  #   x: A single character string to be split on its colon
  #       "Court Date: 10/4/2012 8:30 AM"
  #
  # Returns:
  #   A single-column data frame
  name  <- sub(" ", ".", tolower(x[1]))
  if (name %in% c("violation.date","court.date")){
      val <- as.POSIXct(x[2], format = "%m/%d/%Y %I:%M %p")
    } else {
      val <- x[2]
    }
  out  <- data.frame(val)
  names(out) <- name
  return(out)
}

#' 
#' 
#' And we include that function in the main one extracting entries from the third table 
#' of the case summary.
#' 
## @knitr extractIdentifiers-def
extractIdentifiers <- function(raw){
  # Parses the identifiers table (no. 3) of LVMC case reports 
  #
  # Args:
  #   ids.raw: An XML node set containing a single HTML table
  #
  # Returns:
  #   A data frame containing:
  #         history.number 
  #         case.number 
  #         citation.number 
  #         violation 
  #         violation.date 
  #         name 
  #         department 
  #         court.date  
  #
  
  # Extract the table containing case identifiers
  ids.raw.df <- readHTMLTable(raw[[3]], stringsAsFactors = FALSE)
  # Split each cell into id:value pairs. Not every case has a citation number so
  # we split on a colon ending a line or followed by a space
  ids.raw.lst <- c(strsplit(ids.raw.df$V1, ": |:$"),
                 strsplit(ids.raw.df$V2[!is.na(ids.raw.df$V2)], ": |:$")
                 )
  # Combine each id:value string vector into a data frame
  ids.pairs <- sapply(ids.raw.lst, pairVectors)
  # Merge the list of data frames and exit
  ids <- data.frame(ids.pairs)
  return(ids)
}

#' 
#' 
#' ## Case Docket
#' This table is a bit more complex in that a lot of the data we're after is encoded in the case activity column as id:value pairs (*e.g.*, `Plea: NOLO` or `Finding: GUILTY`). We could repeat the steps we used to extract the `ids` table, but there's additional data that doesn't seem to obey the colon delimited format, most importantly the type of attorney (`Public Attorney SMITH, BRIAN Bar# 11279`). 
#' 
#' | Activity entry                                                | Variables                                                                              | Type         |
#' |---------------------------------------------------------------+----------------------------------------------------------------------------------------+--------------|
#' | Case Closed                                                   | `case.closed = TRUE`                                                                   | logical      |
#' | Plea: NOLO)                                                   | `plea=NC` or (`NG, GT, SU`)                                                                | factor       |
#' | Finding: GUILTY                                               | `finding=GT`or (`NG`, `DS`?)                                                           | factor       |
#' | Probable Cause Found                                          | `pc.hearing=TRUE, pc.found=TRUE`                                                       | logical      |
#' | Bail Due: $1130                                               | `bail due=1130`                                                                        | numeric      |
#' | Sentence: JAIL 3 days CTS: 3 days                             | `list("sentence.type"=J, "sentence.val"=3, "sentence.suspended"=FALSE), jail.served=3` | list,numeric |
#' | Complaint Filed 12/2/2010 9:10 AM                             | `complaint.date=date`                                                                  | POSIXct      |
#' | Public Attorney SMITH, BRIAN Bar# 11279                       | `attorney.type=(PDR,PVT), attorney.name="SMITH, BRIAN",attorney.bar=11279`           | factor       |
#' | BW Ordered                                                    | `bw.issued=TRUE`                                                                       | logical      |
#' | Suspend Sentence for 1Y `AND` Sentence: Suspend JAIL 179 days | ``list("sentence.type"=J, "sentence.val"=3, "sentence.suspended"=TRUE)`                | list         |
#' 
#' There's a separate table for sentences, so we'll set that aside for the moment. We begin by establishing whether the case is closed, and whether the court conducted a [probable cause hearing](http://en.wikipedia.org/wiki/Probable_cause#Probable_cause_hearings): 
#' 
## @knitr miscFunctions-def
isClosed <- function(docket.df) {
  if (length(grep("Case Closed", docket.df$activity,  fixed = TRUE)) == 0) {
    case.closed  <- FALSE
  } else {
    case.closed  <- TRUE
  }
  return(case.closed)
}

GetProbableCause <- function(docket.df) {
  if (length(grep("Probable Cause", docket.df$activity,  fixed = FALSE,
                  ignore.case = TRUE)) == 0) {
    pc.hearing.held  <- FALSE
    pc.found <- NA
  } else {
    pc.hearing.held  <- TRUE
    if (length(grep("Probable Cause Found", docket.df$activity,  fixed = FALSE,
                    ignore.case = TRUE)) > 0){
      pc.found <- TRUE
    } else {
      pc.found <- FALSE
    }
  }
  pc.hearing <- data.frame(pc.hearing.held, pc.found)
  return(pc.hearing)
}

#' 
#' 
#' Getting the attorney name and type requires extracting it via regex. We also have to consider
#' that multiple attorney entries may exist, as when a private attorney is replaced by a publicly provided one. We can do this by setting up the function to only record the first (most recent) entry, and to set certain logical flags if there are multiple attorneys.
#' 
## @knitr parseAttorneyName+getAttorney-def
parseAttorneyName <- function(atty.str){   
    atty.str <- sub("(Public|Private) Attorney ([A-Za-z]+, [A-Za-z]+|[A-Za-z]+, [A-Za-z]+ [A-Za-z]\\.) Bar# (\\d+)", "\\2~\\3", atty.str)
    atty.entry <- unlist(strsplit(atty.str, "~", fixed = TRUE))
    a <- list(attorney.name = atty.entry[1], attorney.bar.num = atty.entry[2])
    if(length(a) == 2) { # check for a nam
      return(a)
      } else {
        a <- list(attorney.name = NA, attorney.bar.num = NA)
        return(a)
      }
}

getAttorney <- function(docket.df) {
  # Parses the case docket for attorney of record entries in LVMC case reports
  # 
  # Args:
  #   docket.df: data frame with date and activity cols
  #
  # Returns:
  #   A data frame containing:
  #         Attorney Name, if mult, first (latest)
  #         attorney bar number, if mult, first (latest)
  #         assigned counsel
  #         multiple attorneys
  
  pri.len <- length(grep("Private Attorney", docket.df$activity,  fixed = TRUE))
  pub.len <- length(grep("Public Attorney", docket.df$activity,  fixed = TRUE))
  if (pri.len == 0 && pub.len == 0) { 
    # No attorney
      atty <- data.frame(attorney.is.assigned = FALSE, 
                         attorney.is.private = FALSE,
                         attorney.name = NA,
                         attorney.bar.num = NA)
      return(atty)
    } else if (pub.len == 1) { 
      # public attorney
    atty.name <- parseAttorneyName(grep("Public Attorney",
                                        docket.df$activity,
                                        value = TRUE,
                                        ignore.case=TRUE)
                                   )
    atty <- data.frame(attorney.is.assigned = TRUE, attorney.is.private = FALSE,
              atty.name, stringsAsFactors = FALSE)
    return(atty)
    } else if (pri.len == 1) { 
      # private attorney
    atty.name <- parseAttorneyName(grep("Private Attorney", docket.df$activity,
                                        value = TRUE, ignore.case=TRUE)
                                   )
    atty <- data.frame(attorney.is.assigned = FALSE,
                       attorney.is.private = TRUE,
                       atty.name,
                       stringsAsFactors = FALSE)
    return(atty)
    } else {      
      atty <- data.frame(attorney.is.assigned = NA, 
                         attorney.is.private = NA,
                         attorney.name = NA,
                         attorney.bar.num = NA)
      return(atty)}
  
  }

#' 
#' 
#' Here we create a simpler version of the `pairVectors` function, which we'll use to extract pleas and findings (verdicts). We also want to check whether the City Attorney declines to pursue charges.
#' 
## @knitr activitySplit+isDenied-def
activitySplit <- function(x) str_trim(unlist(strsplit(x, ": |:$")))

isDenied <- function(docket.df) {
  if (length(grep("(Charge Denied|(Case|Charge) dismissed)", docket.df$activity, ignore.case = TRUE)) == 0) {
    charge.denied  <- FALSE
  } else {
    charge.denied  <- TRUE
  }
  return(charge.denied)
}

#' 
#' 
#' Putting it all together, we have:
#' 
## @knitr parseDocket-def
parseDocket <- function(docket.df) {
  charge.denied <- isDenied(docket.df)
  if (charge.denied) {
    finding <- plea <- "NONE"
    } else {
  plea <- activitySplit(grep("Plea", docket.df$activity, value=TRUE))[2]
  finding <- activitySplit(grep("Finding", docket.df$activity, value=TRUE))[2]
  try(if (plea == "Plea" || finding == "Finding") {
    plea <- activitySplit(grep("Plea", docket.df$activity, value=TRUE))[3]
  finding <- activitySplit(grep("Finding", docket.df$activity, value=TRUE))[3]
  })
  }
  attorney <- getAttorney(docket.df)
  pc.hearing <- GetProbableCause(docket.df)
  dkt <- as.data.frame(list(charge.denied = charge.denied, plea = plea, finding = finding))
  dkt <- cbind(dkt, attorney, pc.hearing)
  return(dkt)
}

#' 
#' 
#' We first detect whether the charges have been denied, and, if so, skip the plea and verdict. Because these are sometimes prefaced by "`Correction:`" we select the third element when we extract the word "plea" or "finding." We then get our attorney data and PC hearing status. We `cbind` these together into a single data frame and return it.
#' 
#' Depending on whether various flags (bench warrants, clerk flags or alerts, etc.) have been set, the location of the docket can vary by up to three slots. But because the record flags are one column HTML tables we can check for the two columns of our
#' docket table ("Activity date" and "Activity"). We set `stringsAsFactors = FALSE` to
#' keep the attorney entries from being returned as a factor and subsequently being
#' converted into an integer when the data frames are combined.
#' 
## @knitr extractDocket-def
extractDocket <- function(raw.tree) {
  docket.raw.df <- readHTMLTable(raw.tree[[5]],
                               stringsAsFactors = FALSE,
                               skip.rows=1,
                               header = TRUE,
                               colClasses = c("caseDate", "character")
                               )
  if (length(docket.raw.df)==2){
    names(docket.raw.df) <- c("activity.date", "activity")
    docket.dt <- as.data.table(docket.raw.df)
    return(docket.dt)
    } else {
    docket.raw.df <- readHTMLTable(raw.tree[[6]],
                               stringsAsFactors = FALSE,
                               skip.rows=1,
                               header = TRUE,
                               colClasses = c("caseDate", "character")
                               )
    if (length(docket.raw.df)==2){
    names(docket.raw.df) <- c("activity.date", "activity")
    docket.dt <- as.data.table(docket.raw.df)
    return(docket.dt)
    } else {
    docket.raw.df <- readHTMLTable(raw.tree[[7]],
                               stringsAsFactors = FALSE,
                               skip.rows=1,
                               header = TRUE,
                               colClasses = c("caseDate", "character")
                               )
    names(docket.raw.df) <- c("activity.date", "activity")
    docket.dt <- as.data.table(docket.raw.df)    
    return(docket.dt)    
    }
    }
  }

#' 
#' 
#' We put all of that together:
#' 
## @knitr ParseSummaryTree-def
ParseSummaryTree <- function(chr) {
  node  <-  TreeParse(chr)
  docket.df <- extractDocket(node)
  case.closed <- isClosed(docket.df)
  if (case.closed) {
  ids <- extractIdentifiers(node)
  dkt <- parseDocket(docket.df)
  case <- cbind(ids, dkt)
  return(case)
  } 
}


#' 
#' 
#' This returns a one row data frame containing:
#' 
#' * `history.number`
#' * `case.number`
#' * `citation.number`
#' * `violation`
#' * `violation.date`
#' * `name`
#' * `department`
#' * `court.date`  
#' * `charge.denied` 
#' * `plea`
#' * `finding`
#' * `attorney.is.assigned`
#' * `attorney.is.private`
#' * `attorney.name`       
#' * `attorney.bar.num`    
#' * `pc.hearing.held`
#' * `pc.found`
#' 
#' ### Extract the dockets
#' 
#' My original approach kept all of the XMLNodeSets in memory, a mistake that quickly filled up 8GB of RAM. 
## @knitr get-summaries, eval=TRUE
library("parallel", lib.loc="/usr/lib/R/library")

if (file.exists("docketsDT.RData")){
  load("docketsDT.RData")  
} else {
system.time(dockets.l <- mclapply(results.list, ParseSummaryTree))
system.time(dockets <- do.call("rbind", dockets.l))
dockets.DT <- data.table(dockets, key = "case.number")
save(dockets.DT, file  = "docketsDT.RData", compress = TRUE)
write.csv(dockets, file = "docketsDT.csv")
}

#' 
#' 
#' A quick glance at the resulting data frame:
#' 
## @knitr summary-table, results='asis'
x.dockets <- xtable(summary(dockets.DT), caption = "Result summaries for the dockets table.")
print(x.dockets, type = "html")

#' 
#' 
#' The most eyepopping result? The 111 acquittals, or `r  prettyNum(111/nrow(dockets.DT) * 100, digits = 3)` percent of `r nrow(dockets.DT)` cases total. That isn't very many at all. A closer look:
#' 
## @knitr dockets-data-table, results='asis'
acquitals <- dockets.DT[finding == 'NOT GUILTY',]
x.acq <- xtable(summary(acquitals), caption = "Result summaries for the acquitals table.")
print(x.acq, type = "html")

#' 
#' 
## @knitr results='asis', echo=FALSE
violdf <- data.frame(dockets.DT[, rev(tail(sort(table(violation)), n=10))])
names(violdf) <- "Count"
x.viol <- xtable(violdf, caption = "Top ten violations for all dockets.")
print(x.viol, type = "html")
acqviol.df <- data.frame(acquitals[, rev(tail(sort(table(violation)), n=10))])
names(acqviol.df) <- "Count"
x.acq.viol <- xtable(acqviol.df, caption = "Top ten violations for dockets of cases resulting in acquittal.")
print(x.acq.viol, type = "html")

#' 
#' 
#' The other interesting result is the `r prettyNum(254/23460 * 100, digits = 3)` percent of probable cause hearings that failed to find probable cause.
#' 
#' 
#' ## Sentencing
#' 
#' Relatively straight forward. We have five columns (sentence name, cash (T/F), due, paid, balance) to extract, and need only check for the presence of a dollar sign in the second column, which indicates a fine.
#' 
## @knitr extractSentence-def
extractSentence <- function(chr, n){
  # Parses the case sentencing table (no. 4) of LVMC case reports 
  #
  # Args:
  #   chr: filename or url pointing to case summary
  #
  # Returns:
  #   A data frame containing:
  #         id
  #         item.name 
  #         cash - todo: character dollar sign -> logical
  #         due
  #         paid
  #         balance
  sen.raw  <-  TreeParse(chr)
  sen.raw.df <- readHTMLTable(sen.raw[[4]], 
                              stringsAsFactors = FALSE,
                              skip.rows=2, 
                              header=c("item",
                                       "cash",
                                       "due",
                                       "paid",
                                       "balance")
                              )
  if (length(sen.raw.df) == 5){
  try(sen.raw.df$cash  <- sen.raw.df$cash %in% "$", silent = TRUE)
  sents <- data.frame(.ids = n, sen.raw.df)
  } else {
    sents <- data.frame(.ids = n,
                        item = NA,                             
                        cash = NA,
                        due = NA,
                        paid = NA,
                        balance = NA
                        )
  }
  return(sents)
}

#' 
#' 
## @knitr get-sentences, cache=F
if (file.exists("sentsDT.RData")){
  load("sentsDT.RData")  
} else {
  sentences.l <- mclapply(seq_along(results.list), function(i) {
    extractSentence(results.list[[i]], n = names(results.list)[[i]])
    }
                          )
  sents <- rbind.fill(sentences.l)
  sents.DT <- data.table(sents, key = ".ids")
  save(sents.DT, file  = "sentsDT.RData", compress = TRUE)
  write.csv(sents.DT, file = "sentsDT.csv")
}

#' 
#' Not evaluated. So the first run:
#' 
#' ```r
#'     > system.time(
#'     + sentences.l <- mclapply(seq_along(results.list), function(i) 
#'     + {
#'     + extractSentence(results.list[[i]], n=names(results.list)[[i]]))
#'     + }
#'     + )
#'        user  system elapsed 
#'     870.088   9.112 436.309 
#'     > 
#'     > 
#'     > system.time(
#'     + sents <- rbind.fill(sentences.l)
#'     + )
#'         user   system  elapsed 
#'     2856.880    3.672 2863.360 
#' ```
#' Which is obviously hosed. It would probably be faster to write each data frame to
#' disk, then paste or `cat` them together than to use `rbind.fill`. The `parallel` package
#' is particularly useful here, cutting processing time in half on my machine (aging i5-650 dual-core CPU).
#' 
## @knitr save-sents
ls()

#' 
#' 
## @knitr tabs
tables()
sessionInfo()

#' 
#' 
#' 
#' 
#' 
#' 
#' 
